import type { ISourceStreamFactory } from "./source.js";
import type { ISinkStreamFactory } from "./sink.js";
import type { IDuplexStreamFactory } from "./duplex.js";
import type { TStreamChunk, TStreamError } from "./stream.js";
interface IPipelineNetworkStream {
    destroy: () => void;
}
interface IPipelineNetworkFactoryStreamArgs {
    done: () => void;
    failed: (args: {
        error: TStreamError;
    }) => void;
}
interface IPipelineNetworkFactory {
    stream: (args: IPipelineNetworkFactoryStreamArgs) => IPipelineNetworkStream;
}
interface IRewireWiring {
    from: ISourceStreamFactory<TStreamChunk>;
    via: IDuplexStreamFactory<TStreamChunk, TStreamChunk>[];
    to: ISinkStreamFactory<TStreamChunk>;
}
interface IRewireablePipelineNetworkFactory extends IPipelineNetworkFactory {
    rewire: (args: {
        wirings: IRewireWiring[];
    }) => void;
}
declare const create: () => IRewireablePipelineNetworkFactory;
declare const createLinear: ({ from, via, to }: {
    from: ISourceStreamFactory<TStreamChunk>;
    via: IDuplexStreamFactory<TStreamChunk, TStreamChunk>[];
    to: ISinkStreamFactory<TStreamChunk>;
}) => IRewireablePipelineNetworkFactory;
type TPipelineSourceOpenFunc = <T extends TStreamChunk>(args: {
    output: ISinkStreamFactory<T>;
}) => IPipelineNetworkFactory;
declare const pipelineSource: <T extends TStreamChunk>({ open }: {
    open: TPipelineSourceOpenFunc;
}) => ISourceStreamFactory<T>;
type TPipelineSinkOpenFunc = (args: {
    input: ISourceStreamFactory<TStreamChunk>;
}) => IPipelineNetworkFactory;
declare const pipelineSink: <T extends TStreamChunk>({ open }: {
    open: TPipelineSinkOpenFunc;
}) => ISinkStreamFactory<T>;
type TPipelineTransformOpenFunc = <T extends TStreamChunk, U extends TStreamChunk>(args: {
    input: ISourceStreamFactory<T>;
    output: ISinkStreamFactory<U>;
}) => IPipelineNetworkFactory;
declare const pipelineTransform: <T extends TStreamChunk, U extends TStreamChunk>({ open }: {
    open: TPipelineTransformOpenFunc;
}) => any;
export { create, createLinear, pipelineSource, pipelineSink, pipelineTransform };
