import { IStreamFactory, IStream, TStreamChunk, TStreamError } from "./stream.js";
interface ISinkStream<T extends TStreamChunk> extends IStream {
    write: ({ chunks }: {
        chunks: T[];
    }) => {
        takesMore: boolean;
    };
    finish: ({ done }: {
        done: () => void;
    }) => void;
}
interface ISinkStreamFactoryOptions {
    drain: () => void;
    fail: ({ error }: {
        error: TStreamError;
    }) => void;
}
interface ISinkStreamFactory<T extends TStreamChunk> extends IStreamFactory {
    open: (options: ISinkStreamFactoryOptions) => ISinkStream<T>;
}
interface IDebuggableSinkStreamFactory<T extends TStreamChunk> extends ISinkStreamFactory<T> {
    debug: (args: {
        log: (args: {
            message: string;
        }) => void;
    }) => ISinkStreamFactory<T>;
}
declare const sink: <T extends TStreamChunk>({ open: providedOpen }: ISinkStreamFactory<T>) => IDebuggableSinkStreamFactory<T>;
declare const syncSink: <T extends TStreamChunk>({ write: providedWrite, finish: providedFinish, destroy: providedDestroy }: {
    write: (args: {
        chunks: T[];
    }) => void;
    finish: () => void;
    destroy: () => void;
}) => IDebuggableSinkStreamFactory<TStreamChunk>;
declare const nullSink: () => IDebuggableSinkStreamFactory<TStreamChunk>;
export type { ISinkStream, ISinkStreamFactory, ISinkStreamFactoryOptions, };
export { sink, syncSink, nullSink };
