import { IStreamFactory, IStream, TStreamChunk, TStreamError } from "./stream.js";
interface ISourceStream<T extends TStreamChunk> extends IStream {
    pause: () => void;
    resume: () => void;
}
interface ISourceStreamFactoryOptions<T extends TStreamChunk> {
    next: ({ chunks }: {
        chunks: T[];
    }) => void;
    end: () => void;
    fail: ({ error }: {
        error: globalThis.Error;
    }) => void;
}
interface ISourceStreamFactory<T extends TStreamChunk> extends IStreamFactory {
    open: (options: ISourceStreamFactoryOptions<T>) => ISourceStream<T>;
}
declare const source: <T extends TStreamChunk>({ open }: ISourceStreamFactory<T>) => ISourceStreamFactory<T>;
declare const errorSource: ({ error }: {
    error: TStreamError;
}) => ISourceStreamFactory<TStreamChunk>;
declare const sourceFromString: ({ data, chunkSize }: {
    data: string;
    chunkSize?: number;
}) => ISourceStreamFactory<string>;
declare const sourceFromChunks: <T extends TStreamChunk>({ chunks }: {
    chunks: T[];
}) => ISourceStreamFactory<T>;
export type { ISourceStream, ISourceStreamFactory, ISourceStreamFactoryOptions };
export { source, errorSource, sourceFromString, sourceFromChunks };
